// TileTiff2.cpp : Defines the entry point for the console application.
//
//
// Copyright (C) 2015 Neuro Programming Research Co., Ltd.
//                    Biomodeling Research Co., Ltd.
//  All rights reserved.
//

#include "stdafx.h"

#include <Windows.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <list>
#include <sstream>
#include <fstream>

// libtiff
#include "tiff\tiffio.h"
#pragma comment(lib,"libtiff.lib")


#pragma pack(push,1)
// tiff header
typedef struct{
	unsigned short Endian;
	unsigned short Version;
	unsigned long IFDoffset;
} TiffHeader;

// tiff DirectoryEntry
typedef struct{
	unsigned short Tag;
	unsigned short Type;
	unsigned long Count;
	union{
		unsigned short ValueShort;
		unsigned long Value;
	};
} DirectoryEntry;
#pragma pack(pop)

// Write a tiff DirectoryEntry to stream
/oid writeEntry(std::ofstream &out, unsigned short Tag, unsigned short Type, unsigned long Count, unsigned long Value)
{
	DirectoryEntry entry = { 0 };
	entry.Tag = Tag;
	entry.Type = Type;
	entry.Count = Count;
	if (Type == 3 && Count == 1)
		entry.ValueShort = (unsigned short)Value;
	else
		entry.Value = Value;
	out.write((const char *)&entry, sizeof(entry));
}

// write tiff file
// compatible with ImageJ(over 4GB tiff stack file)
//
// usage: call constracter and write a file with WriteTiff()
class TiffWriter
{
public:
	TiffWriter(std::wstring filename, WORD *image, int width, int height, int stack);
	void WriteTiff();

private:
	std::ofstream outfile;
	std::string description;
	WORD *image;
	int width;
	int height;
	int stack;
	unsigned long imageSize;
	unsigned long descriptionOffset;
	unsigned long offset;
	const WORD NumberOfDirectoryEntry = 10;
	const int IFDsize = 2 + NumberOfDirectoryEntry * sizeof(DirectoryEntry) + 4;

	void WriteHeader();
	void writeIFD(unsigned long imageOffset, unsigned long nextIFD);
};

// constructor
// TiffWriter(std::wstring filename, WORD *image, int width, int height, int stack)
// filename	:	filename
// image	:	pointer to image data
// width	:	width of image
// height	:	height of image
// stack	:	pages of stack image
TiffWriter::TiffWriter(std::wstring filename, WORD *image, int width, int height, int stack)
{
	this->image = image;
	this->height = height;
	this->width = width;
	this->stack = stack;
	imageSize = sizeof(WORD) * width * height;
	offset = 0;
	outfile.open(filename, std::ios::binary);
	{
		std::stringstream ss;
		ss << "ImageJ=(null)\n"
			<< "images=" << stack << "\n"
			<< "slices=" << stack << "\n";
		description = ss.str();
	}
}

// write tiff header
void TiffWriter::WriteHeader()
{
	TiffHeader header;
	header.Endian = 0x4949; //little endian
	header.Version = 42;
	header.IFDoffset = 8;
	outfile.write((const char *)&header, sizeof(header));
	offset += sizeof(header);
}

// write IFD(Image File Directory)
// output the minimum items needed
void TiffWriter::writeIFD(unsigned long imageOffset, unsigned long nextIFD)
{
	offset += IFDsize;
	outfile.write((const char *)&NumberOfDirectoryEntry, sizeof(NumberOfDirectoryEntry));
	writeEntry(outfile, 0x00fe, 4, 1, 0); //NEW_SUBFILE_TYPE
	writeEntry(outfile, 0x0100, 4, 1, width); //IMAGE_WIDTH
	writeEntry(outfile, 0x0101, 4, 1, height); //IMAGE_LENGTH
	writeEntry(outfile, 0x0102, 3, 1, 16); //BITS_PER_SAMPLE
	writeEntry(outfile, 0x0106, 3, 1, 1); //PHOTO_INTERP
	writeEntry(outfile, 0x010e, 2, (unsigned long)description.length(), descriptionOffset); //IMAGE_DESCRIPTION
	writeEntry(outfile, 0x0111, 4, 1, imageOffset); //STRIP_OFFSETS
	writeEntry(outfile, 0x0115, 3, 1, 1); //SAMPLES_PER_PIXEL
	writeEntry(outfile, 0x0116, 3, 1, height); //ROWS_PER_STRIP
	writeEntry(outfile, 0x0117, 4, 1, imageSize); //STRIP_BYTE_COUNT
	outfile.write((const char *)&nextIFD, sizeof(nextIFD));
}

// write tiff file
void TiffWriter::WriteTiff()
{
	// write a tiff header
	WriteHeader();
	// write description body just after first page IFD
	descriptionOffset = offset + IFDsize;
	// write image body just after description body
	unsigned long imageOffset = descriptionOffset + (unsigned long)description.length();
	// pointer to next IFD
	unsigned long nextIFD = 0;
	// Write IFDs for page 2- at the end of tiff file.
        // If that point is inside 4GB, write it.
	// But if that point is over 4GB, not write.
	// Because, next IFD pointer cannot point over 4GB, no way to point IFD for page 2-.
	// In ImageJ, treat thease files as raw image data file with setting on header.
	// In other application, can read just first page.
	if ((imageOffset + (int64)imageSize*stack + IFDsize*stack) < 0xFFFFFFFFL){
		nextIFD = imageOffset + imageSize*stack;
	}
	// write first page IFD
	writeIFD(imageOffset, nextIFD);
	// write description body
	outfile.write(description.c_str(), description.length());
	offset += (unsigned long)description.length();
	// write image body
	outfile.write((const char *)image, (int64)imageSize*stack);
	// If this point insige 4GB, ...
	if (nextIFD > 0){
		offset += imageSize*stack;
		// write remain IFDs
		for (int i = stack-2; i >= 0; i--){
			nextIFD = (i > 0)?offset + IFDsize: 0;
			imageOffset += imageSize;
			writeIFD(imageOffset, nextIFD);
		}
	}
}

int _tmain(int argc, _TCHAR* argv[])
{
	// usage
	if (argc < 3){
		printf("TileTiff2 (outputfile) (inputdir) [x_size] [y_size] [overlap]\n");
		return 1;
	}

	// 1st arg is output filename
	std::wstring outfile(argv[1]);

	// 2nd arg is input directory name
	std::wstring dirname(argv[2]);
	// end with \
	if (dirname[dirname.length() - 1] != '\\')
		dirname += '\\';

	int x_size = 0;
	int y_size = 0;
	int overlap = 0;
	// not spetify "x_size y_size overlap", read from txt file of summary file
	if (argc < 6){
		WIN32_FIND_DATA fd;
		// find txt file in input directory
		HANDLE hdir = FindFirstFile((dirname + L"*.txt").c_str(), &fd);
		if (hdir != INVALID_HANDLE_VALUE){
			// to full path
			std::wstring fullpath(dirname);
			fullpath += fd.cFileName;
			// open txt file
			std::ifstream txtinfo(fullpath);
			std::string str;
			while (getline(txtinfo, str)){
				// read x_size and y_size
				if (str.find("Montage Positions - ") != std::string::npos){
					sscanf_s(str.c_str(), "%*[^(](%*d%*[^(]( %d by %d)", &x_size, &y_size);
				}
				// read overlap
				if (str.find("Montage=") != std::string::npos){
					sscanf_s(str.c_str(), "%*[^O]Overlap %d", &overlap);
				}
			}
			FindClose(hdir);
		}
	}
	else {
		// specify in 3-5th arg, read them
		std::wstringstream(argv[3]) >> x_size;
		std::wstringstream(argv[4]) >> y_size;
		std::wstringstream(argv[5]) >> overlap;
	}
	printf("X=%d, Y=%d, overlap=%d\n", x_size, y_size, overlap);

	std::list<std::wstring> filelist;
	{
		// search tif files in input directory
		WIN32_FIND_DATA fd;
		HANDLE hdir = FindFirstFile((dirname+L"*.tif*").c_str(), &fd);
		if (hdir != INVALID_HANDLE_VALUE){
			do {
				// add as fullpath
				std::wstring fullpath(dirname);
				fullpath += fd.cFileName;
				filelist.push_back(fullpath);
			} while (FindNextFile(hdir, &fd));
			FindClose(hdir);
		}
	}
	// sort file list
	filelist.sort();

	// suppress the warning in read
	TIFFErrorHandler oldhandler;
	oldhandler = TIFFSetWarningHandler(NULL);

	int StackCount = 0;
	WORD *outpixel = NULL;
	int64 outsize = 0;
	int bigx = 0;
	int bigy = 0;
	int width = 0;
	int height = 0;
	// read a tiff file from list
	for (auto filename : filelist){
		wprintf(L"process...%s\n", filename.c_str());

		// open a tiff file
		TIFF *in = TIFFOpenW(filename.c_str(), "r");
		if (!in){
			printf("error on open tiff\n");
			return 2;
		}
		// if not set stack page count yet, count it.
		if (StackCount == 0){
			do {
				StackCount++;
			} while (TIFFReadDirectory(in));
			TIFFSetDirectory(in, 0);
		}
		// process for every page
		int stack = 0;
		do {
			int nImageLength = 0;
			int nImageWidth = 0;
			int nBitsPerSample = 1;
			int nSamplePerPixel = 1;

			// read tag in needed
			// width, height, bits for pixel, samples for pixel(color)
			if (!TIFFGetField(in, TIFFTAG_IMAGEWIDTH, &nImageWidth) ||
				!TIFFGetField(in, TIFFTAG_IMAGELENGTH, &nImageLength) ||
				!TIFFGetField(in, TIFFTAG_BITSPERSAMPLE, &nBitsPerSample) ||
				!TIFFGetField(in, TIFFTAG_SAMPLESPERPIXEL, &nSamplePerPixel))
			{
				printf("error in tiff tag\n");
				return 1;
			}

			// this app for 16bit monochrome
			if (nBitsPerSample != 16){
				printf("unsupported BitsPerSample %d\n", nBitsPerSample);
				return 1;
			}
			if (nSamplePerPixel != 1){
				printf("unsupported SamplePerPixel %d\n", nSamplePerPixel);
				return 1;
			}

			// if not allocate buffer yet, do it.
			if (!outpixel){
				// size of fusioned image
				width = (int)(nImageWidth * (100 - overlap) / 100) * (x_size - 1) + nImageWidth;
				height = (int)(nImageLength * (100 - overlap) / 100) * (y_size - 1) + nImageLength;
				outsize = (int64)width*height*StackCount;
				try{
					// zero clear
					outpixel = new WORD[outsize]();
				}
				catch (std::bad_alloc) {
					fprintf(stderr, "memory allocation error.\n");
					return 3;
				}
			}

			// allocate 1 line buffer to read
			WORD *buf = (WORD *)_TIFFmalloc(TIFFScanlineSize(in));
			if (!buf){
				fprintf(stderr, "memory allocation error.\n");
				return 3;
			}

			for (int y = 0; y < nImageLength; y++){
				// read one horizontal line
				TIFFReadScanline(in, buf, y);
				for (int x = 0; x < nImageWidth; x++){
					// input tiling pixel value
					WORD value1 = buf[x];
					// output position of output image
					//int outx = (int)((100 - overlap) * (nImageWidth) / 100) * bigx;
					//int outy = (int)((100 - overlap) * (nImageLength) / 100) * bigy;
					int outx = (int)((100 - overlap) * (nImageWidth) / 100) * (x_size - 1 - bigx);
					int outy = (int)((100 - overlap) * (nImageLength) / 100) * (y_size - 1 - bigy);
					int64 index = (int64)stack*height*width + (width - 1 - outx - x) * height + outy + y; //turn left 90
					// read previous value for overlap
					WORD value2 = outpixel[index];
					// max intensity
					outpixel[index] = max(value1, value2);
				}
			}
			_TIFFfree(buf);
			stack++;
		} while (TIFFReadDirectory(in));
		TIFFClose(in);

		// increment tiling potision
		if (++bigx >= x_size){
			bigy++;
			bigx = 0;
		}
		// it's end
		if (bigy >= y_size){
			break;
		}
	}
	// recover the warning
	TIFFSetWarningHandler(oldhandler);

	{
		// output the fusion image to tiff file
		TiffWriter tiff(outfile, outpixel, height, width, StackCount); //turn left 90
		tiff.WriteTiff();
	}
	delete[] outpixel;

	return 0;
}

